"""
A module that contains links to audio files that can be used for test purposes.

.. autosummary::
    :toctree: generated/

    one_to_six_arabic_16000_mono_bc_noise
    was_der_mensch_saet_mono_44100_lead_trail_silence
"""

import os

__all__ = [
    "one_to_six_arabic_16000_mono_bc_noise",
    "was_der_mensch_saet_mono_44100_lead_trail_silence",
]

_current_dir = os.path.dirname(os.path.realpath(__file__))

one_to_six_arabic_16000_mono_bc_noise = "{cd}{sep}data{sep}1to6arabic_\
16000_mono_bc_noise.wav".format(
    cd=_current_dir, sep=os.path.sep
)
"""A wave file that contains a pronunciation of Arabic numbers from 1 to 6"""

was_der_mensch_saet_mono_44100_lead_trail_silence = "{cd}{sep}data{sep}was_\
der_mensch_saet_das_wird_er_vielfach_ernten_44100Hz_mono_lead_trail_\
silence.wav".format(
    cd=_current_dir, sep=os.path.sep
)
"""A wave file that contains a sentence with a long leading and trailing silence"""



"""
:author:

Amine SEHILI <amine.sehili@gmail.com>
2015-2024

:License:

This package is published under the MIT license.
"""

from .core import *
from .exceptions import *
from .io import *
from .util import *

__version__ = "0.3.0"



#!/usr/bin/env python
# encoding: utf-8
"""`auditok` -- An Audio Activity Detection Tool

`auditok` is a program designed for audio or acoustic activity detection.
It supports reading audio data from various sources, including audio files,
microphones, and standard input.

@author:     Mohamed El Amine SEHILI
@copyright:  2015-2024 Mohamed El Amine SEHILI
@license:    MIT
@contact:    amine.sehili@gmail.com
@deffield    updated: 01 Nov 2024
"""

import os
import sys
import threading
import time
from argparse import ArgumentParser

from auditok import AudioRegion, __version__

from .cmdline_util import initialize_workers, make_kwargs, make_logger
from .exceptions import ArgumentError, EndOfProcessing

__all__ = []
__date__ = "2015-11-23"
__updated__ = "2024-10-30"


def main(argv=None):
    program_name = os.path.basename(sys.argv[0])
    if argv is None:
        argv = sys.argv[1:]
    try:
        parser = ArgumentParser(
            prog=program_name,
            description="auditok, an audio tokenization tool.",
        )
        parser.add_argument(
            "--version", "-v", action="version", version=__version__
        )
        group = parser.add_argument_group("Input-Output options:")
        group.add_argument(
            dest="input",
            help="Input audio or video file. Use '-' for stdin "
            "[Default: read from a microphone using PyAudio].",
            metavar="input",
            nargs="?",
            default=None,
        )
        group.add_argument(
            "-I",
            "--input-device-index",
            dest="input_device_index",
            help="Audio device index [Default: %(default)s]. "
            "Optional and only effective when using PyAudio.",
            type=int,
            default=None,
            metavar="INT",
        )
        group.add_argument(
            "-F",
            "--audio-frame-per-buffer",
            dest="frame_per_buffer",
            help="Audio frame per buffer [Default: %(default)s]. "
            "Optional and only effective when using PyAudio.",
            type=int,
            default=1024,
            metavar="INT",
        )
        group.add_argument(
            "-f",
            "--input-format",
            dest="input_format",
            type=str,
            default=None,
            help="Specify the input audio file format. If not provided, the "
            "format is inferred from the file extension. If the output file "
            "name lacks an extension, the format is guessed from the file "
            "header (requires pydub). If neither condition is met, an error "
            "is raised.",
            metavar="STRING",
        )
        group.add_argument(
            "-M",
            "--max-read",
            dest="max_read",
            type=float,
            default=None,
            help="Maximum data (in seconds) to read from a microphone or a file"
            " [Default: read until the end of the file or stream].",
            metavar="FLOAT",
        )
        group.add_argument(
            "-L",
            "--large-file",
            dest="large_file",
            action="store_true",
            default=False,
            help="Whether the input file should be treated as a large file. "
            "If True, data will be read from file on demand, otherwise all "
            "audio data is loaded into memory before tokenization.",
        )
        group.add_argument(
            "-O",
            "--save-stream",
            dest="save_stream",
            type=str,
            default=None,
            help="Save read audio data (from a file or a microphone) to a file."
            " If omitted, no audio data will be saved.",
            metavar="FILE",
        )
        group.add_argument(
            "-o",
            "--save-detections-as",
            dest="save_detections_as",
            type=str,
            default=None,
            help="Specify the file name format to save detected events. You can "
            "use the following placeholders to construct the output file name: "
            "{id} (sequential, starting from 1), {start}, {end}, and {duration}. "
            "Time placeholders are in seconds. "
            "Example: 'Event_{id}{start}-{end}{duration:.3f}.wav'",
            metavar="STRING",
        )
        group.add_argument(
            "-j",
            "--join-detections",
            dest="join_detections",
            type=float,
            default=None,
            help="Join (glue) detected audio events with a specified duration "
            "of silence between them. To be used in combination with the "
            "--save-stream / -O option.",
            metavar="FLOAT",
        )
        group.add_argument(
            "-T",
            "--output-format",
            dest="output_format",
            type=str,
            default=None,
            help="Specify the audio format for saving detections and/or the "
            "main stream. If not provided, the format will be (1) inferred from"
            " the file extension or (2) default to raw format.",
            metavar="STRING",
        )
        group.add_argument(
            "-u",
            "--use-channel",
            dest="use_channel",
            type=str,
            default=None,
            help="Specify the audio channel to use for tokenization when the "
            "input stream is multi-channel (0 refers to the first channel). By "
            "default, this is set to None, meaning all channels are used, "
            "capturing any valid audio event from any channel. Alternatively, "
            "set this to 'mix' (or 'avg'/'average') to combine all channels "
            "into a single averaged channel for tokenization. Regardless of the"
            "option chosen, saved audio events will have the same number of "
            "channels as the input stream. [Default: %(default)s, use all "
            "channels].",
            metavar="INT/STRING",
        )

        group = parser.add_argument_group(
            "Tokenization options:",
            "Set audio events' duration and set the threshold for detection.",
        )
        group.add_argument(
            "-a",
            "--analysis-window",
            dest="analysis_window",
            default=0.01,
            type=float,
            help="Specify the size of the analysis window in seconds. "
            "[Default: %(default)s (10ms)].",
            metavar="FLOAT",
        )
        group.add_argument(
            "-n",
            "--min-duration",
            dest="min_duration",
            type=float,
            default=0.2,
            help="Minimum duration of a valid audio event in seconds. "
            "[Default: %(default)s].",
            metavar="FLOAT",
        )
        group.add_argument(
            "-m",
            "--max-duration",
            dest="max_duration",
            type=float,
            default=5,
            help="Maximum duration of a valid audio event in seconds. "
            "[Default: %(default)s].",
            metavar="FLOAT",
        )
        group.add_argument(
            "-s",
            "--max-silence",
            dest="max_silence",
            type=float,
            default=0.3,
            help="Maximum duration of consecutive silence allowed within a "
            "valid audio event in seconds. [Default: %(default)s]",
            metavar="FLOAT",
        )
        group.add_argument(
            "-d",
            "--drop-trailing-silence",
            dest="drop_trailing_silence",
            action="store_true",
            default=False,
            help="Remove trailing silence from a detection. [Default: trailing "
            "silence is retained].",
        )
        group.add_argument(
            "-R",
            "--strict-min-duration",
            dest="strict_min_duration",
            action="store_true",
            default=False,
            help="Reject events shorter than --min-duration, even if adjacent "
            "to the most recent valid event that reached max-duration. "
            "[Default: retain such events].",
        )
        group.add_argument(
            "-e",
            "--energy-threshold",
            dest="energy_threshold",
            type=float,
            default=50,
            help="Set the log energy threshold for detection. "
            "[Default: %(default)s]",
            metavar="FLOAT",
        )

        group = parser.add_argument_group(
            "Audio parameters:",
            "Set audio parameters when reading from a headerless file "
            "(raw or stdin) or when using custom microphone settings.",
        )
        group.add_argument(
            "-r",
            "--rate",
            dest="sampling_rate",
            type=int,
            default=16000,
            help="Sampling rate of audio data [Default: %(default)s].",
            metavar="INT",
        )
        group.add_argument(
            "-c",
            "--channels",
            dest="channels",
            type=int,
            default=1,
            help="Number of channels of audio data [Default: %(default)s].",
            metavar="INT",
        )
        group.add_argument(
            "-w",
            "--width",
            dest="sample_width",
            type=int,
            default=2,
            help="Number of bytes per audio sample [Default: %(default)s].",
            metavar="INT",
        )

        group = parser.add_argument_group(
            "Use audio events:",
            "Use these options to print, play, or plot detected audio events.",
        )
        group.add_argument(
            "-C",
            "--command",
            dest="command",
            type=str,
            help="Provide a command to execute when an audio event is detected."
            " Use '{file}' as a placeholder for the temporary WAV file "
            "containing the event data (e.g., `-C 'du -h {file}'` to "
            "display the file size or `-C 'play -q {file}'` to play audio "
            "with sox).",
            metavar="STRING",
        )
        group.add_argument(
            "-E",
            "--echo",
            dest="echo",
            action="store_true",
            default=False,
            help="Immediately play back a detected audio event using pyaudio.",
        )
        group.add_argument(
            "-B",
            "--progress-bar",
            dest="progress_bar",
            action="store_true",
            default=False,
            help="Show a progress bar when playing audio.",
        )
        group.add_argument(
            "-p",
            "--plot",
            dest="plot",
            action="store_true",
            default=False,
            help="Plot and displays the audio signal along with detections "
            "(requires matplotlib).",
        )
        group.add_argument(
            "--save-image",
            dest="save_image",
            type=str,
            help="Save the plotted audio signal and detections as a picture "
            "or a PDF file (requires matplotlib).",
            metavar="FILE",
        )
        group.add_argument(
            "--printf",
            dest="printf",
            type=str,
            default="{id} {start} {end}",
            help="Prints information about each audio event on a new line "
            "using the specified format. The format can include text and "
            "placeholders: {id} (sequential, starting from 1), {start}, "
            "{end}, {duration}, and {timestamp}. The first three time "
            "placeholders are in seconds, with formatting controlled by the "
            "--time-format argument. {timestamp} represents the system date "
            "and time of the event, configurable with the --timestamp-format "
            "argument. Example: '[{id}]: {start} -> {end} -- {timestamp}'.",
            metavar="STRING",
        )
        group.add_argument(
            "--time-format",
            dest="time_format",
            type=str,
            default="%S",
            help="Specify the format for printing {start}, {end}, and "
            "{duration} placeholders with --printf. [Default: %(default)s]. "
            "Accepted formats are\n:"
            " - %%S: absolute time in seconds\n"
            " - %%I: absolute time in milliseconds\n"
            " - %%h, %%m, %%s, %%i: converts time into hours, minutes, seconds,"
            " and milliseconds (e.g., %%h:%%m:%%s.%%i) and only displays "
            "provided fields.\nNote that %%S and %%I can only be used "
            "independently.",
            metavar="STRING",
        )
        group.add_argument(
            "--timestamp-format",
            dest="timestamp_format",
            type=str,
            default="%Y/%m/%d %H:%M:%S",
            help="Specify the format used for printing {timestamp}. Should be "
            "a format accepted by the 'datetime' standard module. [Default: "
            "'%%Y/%%m/%%d %%H:%%M:%%S'].",
        )
        parser.add_argument(
            "-q",
            "--quiet",
            dest="quiet",
            action="store_true",
            default=False,
            help="Quiet mode: Do not display any information on the screen.",
        )
        parser.add_argument(
            "-D",
            "--debug",
            dest="debug",
            action="store_true",
            default=False,
            help="Debug mode: output processing operations to STDOUT.",
        )
        parser.add_argument(
            "--debug-file",
            dest="debug_file",
            type=str,
            default=None,
            help="Save processing operations to the specified file.",
            metavar="FILE",
        )

        args = parser.parse_args(argv)
        try:
            kwargs = make_kwargs(args)
        except ArgumentError as exc:
            print(exc, file=sys.stderr)
            return 1

        logger = make_logger(args.debug, args.debug_file)

        stream_saver, tokenizer_worker = initialize_workers(
            logger=logger, **kwargs.split, **kwargs.io, **kwargs.miscellaneous
        )
        tokenizer_worker.start_all()

        while True:
            time.sleep(1)
            if len(threading.enumerate()) == 1:
                raise EndOfProcessing

    except (KeyboardInterrupt, EndOfProcessing):
        if tokenizer_worker is not None:
            tokenizer_worker.stop_all()

            if stream_saver is not None:
                stream_saver.join()
                try:
                    stream_saver.export_audio()
                except Exception as aee:
                    print(aee, file=sys.stderr)

            if args.plot or args.save_image is not None:
                from .plotting import plot

                reader = tokenizer_worker.reader

                reader.rewind()
                record = AudioRegion(
                    reader.data, reader.sr, reader.sw, reader.ch
                )
                detections = (
                    (det.start, det.end) for det in tokenizer_worker.detections
                )
                plot(
                    record,
                    detections=detections,
                    energy_threshold=args.energy_threshold,
                    show=True,
                    save_as=args.save_image,
                )
        return 0


if __name__ == "__main__":
    sys.exit(main(None))



"""
Module for main data structures and tokenization algorithms.

.. autosummary::
    :toctree: generated/

    load
    split
    make_silence
    split_and_join_with_silence
    AudioRegion
    StreamTokenizer
"""

import math
import os
import warnings
from dataclasses import dataclass, field
from pathlib import Path

from .exceptions import AudioParameterError, TooSmallBlockDuration
from .io import check_audio_data, get_audio_source, player_for, to_file
from .plotting import plot
from .util import AudioEnergyValidator, AudioReader, DataValidator

try:
    from . import signal_numpy as signal
except ImportError:
    from . import signal

__all__ = [
    "load",
    "split",
    "make_silence",
    "split_and_join_with_silence",
    "AudioRegion",
    "StreamTokenizer",
]


DEFAULT_ANALYSIS_WINDOW = 0.05
DEFAULT_ENERGY_THRESHOLD = 50
_EPSILON = 1e-10


def load(input, skip=0, max_read=None, **kwargs):
    """
    Load audio data from a specified source and return it as an
    :class:`AudioRegion`.

    Parameters
    ----------
    input : None, str, Path, bytes, AudioSource
        The source from which to read audio data. If a `str` or `Path`, it
        should specify the path to a valid audio file. If `bytes`, it is
        treated as raw audio data. If set to "-", raw data will be read from
        standard input (stdin). If `None`, audio data is read from the
        microphone using PyAudio. For `bytes` data or a raw audio file path,
        `sampling_rate`, `sample_width`, and `channels` parameters (or their
        aliases) must be specified. If an :class:`AudioSource` object is
        provided, it is used directly to read data.
    skip : float, default: 0
        Duration in seconds of audio data to skip from the beginning of the
        source. When reading from a microphone, `skip` must be 0; otherwise,
        a `ValueError` is raised.
    max_read : float, default: None
        Duration in seconds of audio data to read from the source. When reading
        from the microphone, `max_read` must not be `None`; otherwise, a
        `ValueError` is raised.
    audio_format, fmt : str
        Format of the audio data (e.g., wav, ogg, flac, raw, etc.). This is
        only used if `input` is a string path to an audio file. If not
        provided, the audio format is inferred from the file's extension or
        header.
    sampling_rate, sr : int
        Sampling rate of the audio data. Required if `input` is a raw audio
        file, a `bytes` object, or `None` (i.e., when reading from the
        microphone).
    sample_width, sw : int
        Number of bytes used to encode a single audio sample, typically 1, 2,
        or 4. Required for raw audio data; see `sampling_rate`.
    channels, ch : int
        Number of channels in the audio data. Required for raw audio data;
        see `sampling_rate`.
    large_file : bool, default: False
        If `True`, and `input` is a path to a *wav* or *raw* audio file, the
        file is not fully loaded into memory to create the region (only the
        necessary portion of data is loaded). This should be set to `True`
        when `max_read` is much smaller than the total size of a large audio
        file, to avoid loading the entire file into memory.

    Returns
    -------
    region : AudioRegion

    Raises
    ------
    ValueError
        Raised if `input` is `None` (i.e., reading from the microphone) and
        `skip` is not 0, or if `max_read` is `None` when `input` is `None`.
        This ensures that when reading from the microphone, no data is
        skipped, and the maximum amount of data to read is explicitly
        specified.
    """

    return AudioRegion.load(input, skip, max_read, **kwargs)


def split(
    input,
    min_dur=0.2,
    max_dur=5,
    max_silence=0.3,
    drop_trailing_silence=False,
    strict_min_dur=False,
    **kwargs,
):
    """
    Split audio data and return a generator of :class:`AudioRegion`s.

    Parameters
    ----------
    input : str, Path, bytes, AudioSource, AudioReader, AudioRegion, or None
        Audio data input. If `str` or `Path`, it should be the path to an audio
        file. Use "-" to indicate standard input. If bytes, the input is treated
        as raw audio data. If None, audio is read from the microphone.

        Any input not of type `AudioReader` is converted into an `AudioReader`
        before processing. If `input` is raw audio data (str, bytes, or None),
        specify audio parameters using kwargs (e.g., `sampling_rate`,
        `sample_width`, `channels`).

        For string inputs, audio format is inferred from the file extension, or
        specify explicitly via `audio_format` or `fmt`. Otherwise, the backend
        (currently only `pydub`) handles loading data.

    min_dur : float, default=0.2
        Minimum duration in seconds of a detected audio event. Higher values
        can exclude very short utterances (e.g., single words like "yes" or
        "no"). Lower values may increase the number of short audio events.

    max_dur : float, default=5
        Maximum duration in seconds for an audio event. Events longer than this
        are truncated. If the remainder of a truncated event is shorter than
        `min_dur`, it is included as a valid event if `strict_min_dur` is False;
        otherwise, it is rejected.

    max_silence : float, default=0.3
        Maximum duration of continuous silence allowed within an audio event.
        Multiple silent gaps of this duration may appear in a single event.
        Trailing silence at the end of an event is kept if
        `drop_trailing_silence` is False.

    drop_trailing_silence : bool, default=False
        Whether to remove trailing silence from detected events. To prevent
        abrupt speech cuts, it is recommended to keep trailing silence, so
        default is False.

    strict_min_dur : bool, default=False
        Whether to strictly enforce `min_dur` for all events, rejecting any
        event shorter than `min_dur`, even if contiguous with a valid event.

    Other Parameters
    ----------------
    analysis_window, aw : float, default=0.05 (50 ms)
        Duration of analysis window in seconds. Values between 0.01 and 0.1 are
        generally effective.

    audio_format, fmt : str
        Type of audio data (e.g., wav, ogg, flac, raw). Used if `input` is a
        file path. If not specified, audio format is inferred from the file
        extension or header.

    sampling_rate, sr : int
        Sampling rate of audio data, required if `input` is raw data (bytes or
        None).

    sample_width, sw : int
        Number of bytes per audio sample (typically 1, 2, or 4). Required for
        raw audio; see `sampling_rate`.

    channels, ch : int
        Number of audio channels. Required for raw data; see `sampling_rate`.

    use_channel, uc : {None, "mix"} or int
        Channel selection for splitting if `input` has multiple channels. All
        channels are retained in detected events. Options:

        - None or "any" (default): accept activity from any channel.
        - "mix" or "average": mix all channels into a single averaged channel.
        - int (0 <= value < channels): use the specified channel ID for splitting.

    large_file : bool, default=False
        If True and `input` is a path to a wav or raw file, audio is processed
        lazily. Otherwise, the entire file is loaded before splitting. Set to
        True if file size exceeds available memory.

    max_read, mr : float, default=None
        Maximum data read from source in seconds. Default is to read to end.

    validator, val : callable or DataValidator, default=None
        Custom validator for audio data. If None, uses `AudioEnergyValidator`
        with the given `energy_threshold`. Should be callable or an instance of
        `DataValidator` implementing `is_valid`.

    energy_threshold, eth : float, default=50
        Energy threshold for audio activity detection. Audio regions with
        sufficient signal energy above this threshold are considered valid.
        Calculated as the log energy: `20 * log10(sqrt(dot(x, x) / len(x)))`.
        Ignored if `validator` is specified.

    Yields
    ------
    AudioRegion
        Generator yielding detected :class:`AudioRegion` instances.
    """

    if min_dur <= 0:
        raise ValueError(f"'min_dur' ({min_dur}) must be > 0")
    if max_dur <= 0:
        raise ValueError(f"'max_dur' ({max_dur}) must be > 0")
    if max_silence < 0:
        raise ValueError(f"'max_silence' ({max_silence}) must be >= 0")

    if isinstance(input, AudioReader):
        source = input
        analysis_window = source.block_dur
    else:
        analysis_window = kwargs.get(
            "analysis_window", kwargs.get("aw", DEFAULT_ANALYSIS_WINDOW)
        )
        if analysis_window <= 0:
            raise ValueError(
                f"'analysis_window' ({analysis_window}) must be > 0"
            )

        params = kwargs.copy()
        params["max_read"] = params.get("max_read", params.get("mr"))
        params["audio_format"] = params.get("audio_format", params.get("fmt"))
        if isinstance(input, AudioRegion):
            params["sampling_rate"] = input.sr
            params["sample_width"] = input.sw
            params["channels"] = input.ch
            input = bytes(input)
        try:
            source = AudioReader(input, block_dur=analysis_window, **params)
        except TooSmallBlockDuration as exc:
            err_msg = f"Too small 'analysis_window' ({exc.block_dur}) for "
            err_msg += f"sampling rate ({exc.sampling_rate}). Analysis window "
            err_msg += f"should at least be 1/{exc.sampling_rate} to cover "
            err_msg += "one data sample"
            raise ValueError(err_msg) from exc

    validator = kwargs.get("validator", kwargs.get("val"))
    if validator is None:
        energy_threshold = kwargs.get(
            "energy_threshold", kwargs.get("eth", DEFAULT_ENERGY_THRESHOLD)
        )
        use_channel = kwargs.get("use_channel", kwargs.get("uc"))
        validator = AudioEnergyValidator(
            energy_threshold, source.sw, source.ch, use_channel=use_channel
        )
    mode = StreamTokenizer.DROP_TRAILING_SILENCE if drop_trailing_silence else 0
    if strict_min_dur:
        mode |= StreamTokenizer.STRICT_MIN_LENGTH
    min_length = _duration_to_nb_windows(min_dur, analysis_window, math.ceil)
    max_length = _duration_to_nb_windows(
        max_dur, analysis_window, math.floor, _EPSILON
    )
    max_continuous_silence = _duration_to_nb_windows(
        max_silence, analysis_window, math.floor, _EPSILON
    )

    err_msg = "({0} sec.) results in {1} analysis window(s) "
    err_msg += "({1} == {6}({0} / {2})) which is {5} the number "
    err_msg += "of analysis window(s) for 'max_dur' ({3} == floor({4} / {2}))"
    if min_length > max_length:
        err_msg = "'min_dur' " + err_msg
        raise ValueError(
            err_msg.format(
                min_dur,
                min_length,
                analysis_window,
                max_length,
                max_dur,
                "higher than",
                "ceil",
            )
        )

    if max_continuous_silence >= max_length:
        err_msg = "'max_silence' " + err_msg
        raise ValueError(
            err_msg.format(
                max_silence,
                max_continuous_silence,
                analysis_window,
                max_length,
                max_dur,
                "higher or equal to",
                "floor",
            )
        )

    tokenizer = StreamTokenizer(
        validator, min_length, max_length, max_continuous_silence, mode=mode
    )
    source.open()
    token_gen = tokenizer.tokenize(source, generator=True)
    region_gen = (
        _make_audio_region(
            token[0],
            token[1],
            source.block_dur,
            source.sr,
            source.sw,
            source.ch,
        )
        for token in token_gen
    )
    return region_gen


def make_silence(duration, sampling_rate=16000, sample_width=2, channels=1):
    """
    Generate a silence of specified duration.

    Parameters
    ----------
    duration : float
        Duration of silence in seconds.
    sampling_rate : int, optional
        Sampling rate of the audio data, default is 16000.
    sample_width : int, optional
        Number of bytes per audio sample, default is 2.
    channels : int, optional
        Number of audio channels, default is 1.

    Returns
    -------
    AudioRegion
        A "silent" AudioRegion of the specified duration.
    """
    size = round(duration * sampling_rate) * sample_width * channels
    data = b"\0" * size
    region = AudioRegion(data, sampling_rate, sample_width, channels)
    return region


def split_and_join_with_silence(input, silence_duration, **kwargs):
    """
    Split input audio and join (glue) the resulting regions with a specified
    silence duration between them. This can be used to adjust the length of
    silence between audio events, either shortening or lengthening pauses.

    Parameters
    ----------
    silence_duration : float
        Duration of silence in seconds between audio events.

    Returns
    -------
    AudioRegion or None
        An :meth:`AudioRegion` with the specified between-events silence
        duration. Returns None if no audio events are detected in the input
        data.
    """
    regions = list(split(input, **kwargs))
    if regions:
        first = regions[0]
        # create a silence with the same parameters as input audio
        silence = make_silence(silence_duration, first.sr, first.sw, first.ch)
        return silence.join(regions)
    return None


def _duration_to_nb_windows(
    duration, analysis_window, round_fn=round, epsilon=0
):
    """
    Helper function to convert a given duration into a positive integer
    of analysis windows. If `duration / analysis_window` is not an integer,
    the result will be rounded up to the nearest integer. If `duration == 0`,
    returns 0. If `duration < analysis_window`, returns 1.

    Both `duration` and `analysis_window` should be in the same units,
    either seconds or milliseconds.

    Parameters
    ----------
    duration : float
        The given duration in seconds or milliseconds.
    analysis_window : float
        The size of each analysis window, in the same units as `duration`.
    round_fn : callable, optional
        A function for rounding the result, default is `round`.
    epsilon : float, optional
        A small value added before rounding to address floating-point
        precision issues, ensuring accurate rounding for cases like
        `0.3 / 0.1`, where `round_fn=math.floor` would otherwise yield
        an incorrect result.

    Returns
    -------
    nb_windows : int
        The minimum number of `analysis_window` units needed to cover
        `duration`, ensuring `analysis_window * nb_windows >= duration`.
    """

    if duration < 0 or analysis_window <= 0:
        err_msg = "'duration' ({}) must be >= 0 and 'analysis_window' ({}) > 0"
        raise ValueError(err_msg.format(duration, analysis_window))
    if duration == 0:
        return 0
    return int(round_fn(duration / analysis_window + epsilon))


def _make_audio_region(
    data_frames,
    start_frame,
    frame_duration,
    sampling_rate,
    sample_width,
    channels,
):
    """
    Helper function to create an :class:`AudioRegion` from parameters provided
    by a tokenization object. This function handles setting the `start` and `end`
    metadata for the region.

    Parameters
    ----------
    frame_duration : float
        Duration of each analysis window in seconds.
    start_frame : int
        Index of the first analysis window.
    sampling_rate : int
        Sampling rate of the audio data.
    sample_width : int
        Number of bytes per audio sample.
    channels : int
        Number of audio channels.

    Returns
    -------
    audio_region : AudioRegion
        An AudioRegion object with `start` time calculated as:
        `1000 * start_frame * frame_duration`.
    """
    start = start_frame * frame_duration
    data = b"".join(data_frames)
    return AudioRegion(data, sampling_rate, sample_width, channels, start)


def _read_chunks_online(max_read, **kwargs):
    """
    Helper function to read audio data from an online blocking source
    (e.g., a microphone). This function builds an `AudioRegion` and can
    intercept `KeyboardInterrupt` to stop reading immediately when the
    exception is raised, making it more user-friendly for [i]Python sessions
    and Jupyter notebooks.

    Parameters
    ----------
    max_read : float
        Maximum duration of audio data to read, in seconds.
    kwargs :
        Additional audio parameters (e.g., `sampling_rate`, `sample_width`,
        and `channels`).
    """
    reader = AudioReader(None, block_dur=0.5, max_read=max_read, **kwargs)
    reader.open()
    data = []
    try:
        while True:
            frame = reader.read()
            if frame is None:
                break
            data.append(frame)
    except KeyboardInterrupt:
        # Stop data acquisition from microphone when pressing
        # Ctrl+C in an [i]python session or a notebook
        pass
    reader.close()
    return (
        b"".join(data),
        reader.sampling_rate,
        reader.sample_width,
        reader.channels,
    )


def _read_offline(input, skip=0, max_read=None, **kwargs):
    """
    Helper function to read audio data from an offline source (e.g., file).
    This function is used to build :class:`AudioRegion` objects.

    Parameters
    ----------
    input : str or bytes
        Path to an audio file (if str) or a bytes object representing raw
        audio data.
    skip : float, optional, default=0
        Amount of data to skip from the beginning of the audio source, in
        seconds.
    max_read : float, optional, default=None
        Maximum duration of audio data to read, in seconds. Default is None,
        which reads until the end of the stream.
    kwargs :
        Additional audio parameters (e.g., `sampling_rate`, `sample_width`,
        and `channels`).
    """

    audio_source = get_audio_source(input, **kwargs)
    audio_source.open()
    if skip is not None and skip > 0:
        skip_samples = round(skip * audio_source.sampling_rate)
        audio_source.read(skip_samples)
    if max_read is not None:
        if max_read < 0:
            max_read = None
        else:
            max_read = round(max_read * audio_source.sampling_rate)
    data = audio_source.read(max_read)
    audio_source.close()
    return (
        data,
        audio_source.sampling_rate,
        audio_source.sample_width,
        audio_source.channels,
    )


def _check_convert_index(index, types, err_msg):
    if not isinstance(index, slice) or index.step is not None:
        raise TypeError(err_msg)
    start = index.start if index.start is not None else 0
    stop = index.stop
    for index in (start, stop):
        if index is not None and not isinstance(index, types):
            raise TypeError(err_msg)
    return start, stop


class _SecondsView:
    """
    A class to create a view of an :class:`AudioRegion` that supports slicing
    with time-based indices in seconds.
    """

    def __init__(self, region):
        self._region = region

    def __getitem__(self, index):
        err_msg = "Slicing AudioRegion by seconds requires indices of type "
        err_msg += "'int' or 'float' without a step (e.g. region.sec[7.5:10])"
        start_s, stop_s = _check_convert_index(index, (int, float), err_msg)
        sr = self._region.sampling_rate
        start_sample = int(start_s * sr)
        stop_sample = None if stop_s is None else round(stop_s * sr)
        return self._region[start_sample:stop_sample]

    @property
    def len(self):
        """
        Return region duration in seconds.
        """
        return self._region.duration


class _MillisView(_SecondsView):
    """A class to create a view of `AudioRegion` that can be sliced using
    indices in milliseconds.
    """

    def __getitem__(self, index):
        err_msg = (
            "Slicing AudioRegion by milliseconds requires indices of type "
        )
        err_msg += "'int' without a step (e.g. region.sec[500:1500])"
        start_ms, stop_ms = _check_convert_index(index, (int), err_msg)
        start_sec = start_ms / 1000
        stop_sec = None if stop_ms is None else stop_ms / 1000
        index = slice(start_sec, stop_sec)
        return super().__getitem__(index)

    def __len__(self):
        """
        Return region duration in milliseconds.
        """
        return round(self._region.duration * 1000)

    @property
    def len(self):
        """
        Return region duration in milliseconds.
        """
        return len(self)


class _AudioRegionMetadata(dict):
    """A class to store :class:`AudioRegion`'s metadata."""

    def __getattr__(self, name):
        warnings.warn(
            "`AudioRegion.meta` is deprecated and will be removed in future "
            "versions. For the 'start' and 'end' fields, please use "
            "`AudioRegion.start` and `AudioRegion.end`.",
            DeprecationWarning,
            stacklevel=2,
        )
        if name in self:
            return self[name]
        else:
            err_msg = "AudioRegion metadata has no entry '{}'"
            raise AttributeError(err_msg.format(name))

    def __setattr__(self, name, value):
        self[name] = value

    def __str__(self):
        return "\n".join("{}: {}".format(k, v) for k, v in self.items())

    def __repr__(self):
        return str(self)


@dataclass(frozen=True)
class AudioRegion(object):
    """
    `AudioRegion` encapsulates raw audio data and provides an interface for
    performing basic audio operations. Use :meth:`AudioRegion.load` or
    :func:`load` to create an `AudioRegion` from various input types.

    Parameters
    ----------
    data : bytes
        Raw audio data as a bytes object.
    sampling_rate : int
        Sampling rate of the audio data.
    sample_width : int
        Number of bytes per audio sample.
    channels : int
        Number of audio channels.
    start : float, optional, default=None
        Optional start time of the region, typically provided by the `split`
        function.
    """

    data: bytes
    sampling_rate: int
    sample_width: int
    channels: int
    start: float = field(default=None, repr=None)

    def __post_init__(self):

        check_audio_data(self.data, self.sample_width, self.channels)
        object.__setattr__(self, "splitp", self.split_and_plot)
        duration = len(self.data) / (
            self.sampling_rate * self.sample_width * self.channels
        )
        object.__setattr__(self, "duration", duration)

        if self.start is not None:
            object.__setattr__(self, "end", self.start + self.duration)
            object.__setattr__(
                self,
                "meta",
                _AudioRegionMetadata({"start": self.start, "end": self.end}),
            )
        else:
            object.__setattr__(self, "end", None)
            object.__setattr__(self, "meta", None)

        # `seconds` and `millis` are defined below as @property with docstring
        object.__setattr__(self, "_seconds_view", _SecondsView(self))
        object.__setattr__(self, "_millis_view", _MillisView(self))

        object.__setattr__(self, "sec", self.seconds)
        object.__setattr__(self, "s", self.seconds)
        object.__setattr__(self, "ms", self.millis)

    @classmethod
    def load(cls, input, skip=0, max_read=None, **kwargs):
        """
        Create an :class:`AudioRegion` by loading data from `input`.

        See :func:`load` for a full description of parameters.

        Returns
        -------
        region : AudioRegion
            An AudioRegion instance created from the specified input data.

        Raises
        ------
        ValueError
            Raised if `input` is None and either `skip` is not 0 or `max_read`
            is None.
        """

        if input is None:
            if skip > 0:
                raise ValueError(
                    "'skip' should be 0 when reading from microphone"
                )
            if max_read is None or max_read < 0:
                raise ValueError(
                    "'max_read' should not be None when reading from "
                    "microphone"
                )
            data, sampling_rate, sample_width, channels = _read_chunks_online(
                max_read, **kwargs
            )
        else:
            data, sampling_rate, sample_width, channels = _read_offline(
                input, skip=skip, max_read=max_read, **kwargs
            )

        return cls(data, sampling_rate, sample_width, channels)

    @property
    def seconds(self):
        """
        A view to slice audio region by seconds using
        ``region.seconds[start:end]``.
        """
        return self._seconds_view

    @property
    def millis(self):
        """A view to slice audio region by milliseconds using
        ``region.millis[start:end]``."""
        return self._millis_view

    @property
    def sr(self):
        """Sampling rate of audio data, alias for `sampling_rate`."""
        return self.sampling_rate

    @property
    def sw(self):
        """Number of bytes per sample, alias for `sample_width`."""
        return self.sample_width

    @property
    def ch(self):
        """Number of channels of audio data, alias for `channels`."""
        return self.channels

    def play(self, progress_bar=False, player=None, **progress_bar_kwargs):
        """
        Play the audio region.

        Parameters
        ----------
        progress_bar : bool, optional, default=False
            Whether to display a progress bar during playback. Requires `tqdm`,
            if not installed, no progress bar will be shown.
        player : AudioPlayer, optional, default=None
            Audio player to use for playback. If None (default), a new player is
            obtained via `player_for()`.
        progress_bar_kwargs : dict, optional
            Additional keyword arguments to pass to the `tqdm` progress bar
            (e.g., `leave=False` to clear the bar from the screen upon completion).
        """

        if player is None:
            player = player_for(self)
        player.play(self.data, progress_bar=progress_bar, **progress_bar_kwargs)

    def save(
        self, filename, audio_format=None, exists_ok=True, **audio_parameters
    ):
        """
        Save the audio region to a file.

        Parameters
        ----------
        filename : str or Path
            Path to the output audio file. If a string, it may include `{start}`,
            `{end}`, and `{duration}` placeholders. Regions created by `split`
            contain `start` and `end` attributes that can be used to format the
            filename, as shown in the example.
        audio_format : str, optional, default=None
            Format used to save the audio data. If None (default), the format is
            inferred from the file extension. If the filename has no extension,
            the audio is saved as a raw (headerless) audio file.
        exists_ok : bool, optional, default=True
            If True, overwrite the file if it already exists. If False, raise an
            `IOError` if the file exists.
        audio_parameters : dict, optional
            Additional keyword arguments to pass to the audio-saving backend.

        Returns
        -------
        file : str
            The output filename with placeholders filled in.

        Raises
        ------
        IOError
            If `filename` exists and `exists_ok` is False.

        Examples
        --------
        Create an AudioRegion, specifying `start`. The `end` will be computed
        based on `start` and the region's duration.

        >>> region = AudioRegion(b'\0' * 2 * 24000,
        >>>                      sampling_rate=16000,
        >>>                      sample_width=2,
        >>>                      channels=1,
        >>>                      start=2.25)
        >>> region
        <AudioRegion(duration=1.500, sampling_rate=16000, sample_width=2, channels=1)>

        >>> assert region.end == 3.75
        >>> assert region.save('audio_{start}-{end}.wav') == "audio_2.25-3.75.wav"
        >>> filename = region.save('audio_{start:.3f}-{end:.3f}_{duration:.3f}.wav')
        >>> assert filename == "audio_2.250-3.750_1.500.wav"
        """

        if isinstance(filename, Path):
            if not exists_ok and filename.exists():
                raise FileExistsError(
                    "file '{filename}' exists".format(filename=str(filename))
                )
        if isinstance(filename, str):
            filename = filename.format(
                duration=self.duration,
                meta=self.meta,
                start=self.start,
                end=self.end,
            )
            if not exists_ok and os.path.exists(filename):
                raise FileExistsError(
                    "file '{filename}' exists".format(filename=filename)
                )
        to_file(
            self.data,
            filename,
            audio_format,
            sr=self.sr,
            sw=self.sw,
            ch=self.ch,
            audio_parameters=audio_parameters,
        )
        return filename

    def split(
        self,
        min_dur=0.2,
        max_dur=5,
        max_silence=0.3,
        drop_trailing_silence=False,
        strict_min_dur=False,
        **kwargs,
    ):
        """
        Split audio region. See :func:`auditok.split` for a comprehensive
        description of split parameters.
        See Also :meth:`AudioRegio.split_and_plot`.
        """
        if kwargs.get("max_read", kwargs.get("mr")) is not None:
            warn_msg = "'max_read' (or 'mr') should not be used with "
            warn_msg += "AudioRegion.split_and_plot(). You should rather "
            warn_msg += "slice audio region before calling this method"
            raise RuntimeWarning(warn_msg)
        return split(
            self,
            min_dur=min_dur,
            max_dur=max_dur,
            max_silence=max_silence,
            drop_trailing_silence=drop_trailing_silence,
            strict_min_dur=strict_min_dur,
            **kwargs,
        )

    def plot(
        self,
        scale_signal=True,
        show=True,
        figsize=None,
        save_as=None,
        dpi=120,
        theme="auditok",
    ):
        """
        Plot the audio region with one subplot per channel.

        Parameters
        ----------
        scale_signal : bool, optional, default=True
            If True, scale the signal by subtracting its mean and dividing by its
            standard deviation before plotting.
        show : bool, optional, default=False
            Whether to display the plot immediately after the function call.
        figsize : tuple, optional, default=None
            Width and height of the figure, passed to `matplotlib`.
        save_as : str, optional, default=None
            If specified, save the plot to the given filename.
        dpi : int, optional, default=120
            Dots per inch (DPI) for the plot, passed to `matplotlib`.
        theme : str or dict, optional, default="auditok"
            Plot theme to use. Only the "auditok" theme is currently implemented.
            To define a custom theme, refer to
            :attr:`auditok.plotting.AUDITOK_PLOT_THEME`.
        """

        plot(
            self,
            scale_signal=scale_signal,
            show=show,
            figsize=figsize,
            save_as=save_as,
            dpi=dpi,
            theme=theme,
        )

    def split_and_plot(
        self,
        min_dur=0.2,
        max_dur=5,
        max_silence=0.3,
        drop_trailing_silence=False,
        strict_min_dur=False,
        scale_signal=True,
        show=True,
        figsize=None,
        save_as=None,
        dpi=120,
        theme="auditok",
        **kwargs,
    ):
        """
        Split the audio region, then plot the signal and detected regions.

        Alias
        -----
        :meth:`splitp`

        Refer to :func:`auditok.split()` for a detailed description of split
        parameters, and to :meth:`plot` for plot-specific parameters.
        """
        regions = self.split(
            min_dur=min_dur,
            max_dur=max_dur,
            max_silence=max_silence,
            drop_trailing_silence=drop_trailing_silence,
            strict_min_dur=strict_min_dur,
            **kwargs,
        )
        regions = list(regions)
        detections = ((reg.meta.start, reg.meta.end) for reg in regions)
        eth = kwargs.get(
            "energy_threshold", kwargs.get("eth", DEFAULT_ENERGY_THRESHOLD)
        )
        plot(
            self,
            scale_signal=scale_signal,
            detections=detections,
            energy_threshold=eth,
            show=show,
            figsize=figsize,
            save_as=save_as,
            dpi=dpi,
            theme=theme,
        )
        return regions

    def _check_other_parameters(self, other):
        if other.sr != self.sr:
            raise AudioParameterError(
                "Can only concatenate AudioRegions of the same "
                "sampling rate ({} != {})".format(self.sr, other.sr)
            )
        if other.sw != self.sw:
            raise AudioParameterError(
                "Can only concatenate AudioRegions of the same "
                "sample width ({} != {})".format(self.sw, other.sw)
            )
        if other.ch != self.ch:
            raise AudioParameterError(
                "Can only concatenate AudioRegions of the same "
                "number of channels ({} != {})".format(self.ch, other.ch)
            )

    def _check_iter_others(self, others):
        for other in others:
            self._check_other_parameters(other)
            yield other

    def join(self, others):
        data = self.data.join(
            other.data for other in self._check_iter_others(others)
        )
        return AudioRegion(data, self.sr, self.sw, self.ch)

    @property
    def samples(self):
        warnings.warn(
            "`AudioRegion.samples` is deprecated and will be removed in future "
            "versions. Please use `AudioRegion.numpy()`.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.numpy()

    def __array__(self):
        return self.numpy()

    def numpy(self):
        """Audio region a 2D numpy array of shape (n_channels, n_samples)."""
        return signal.to_array(self.data, self.sample_width, self.channels)

    def __len__(self):
        """
        Return region length in number of samples.
        """
        return len(self.data) // (self.sample_width * self.channels)

    @property
    def len(self):
        """
        Return the length of the audio region in number of samples.
        """

        return len(self)

    def __bytes__(self):
        return self.data

    def __str__(self):
        return (
            "AudioRegion(duration={:.3f}, "
            "sampling_rate={}, sample_width={}, channels={})".format(
                self.duration, self.sr, self.sw, self.ch
            )
        )

    def __repr__(self):
        return "<{}>".format(str(self))

    def __add__(self, other):
        """
        Concatenate this audio region with `other`, returning a new region.

        Both regions must have the same sampling rate, sample width, and number
        of channels. If they differ, a `ValueError` is raised.
        """

        if not isinstance(other, AudioRegion):
            raise TypeError(
                "Can only concatenate AudioRegion, "
                'not "{}"'.format(type(other))
            )
        self._check_other_parameters(other)
        data = self.data + other.data
        return AudioRegion(data, self.sr, self.sw, self.ch)

    def __radd__(self, other):
        """
        Concatenate `other` with this audio region.

        Parameters
        ----------
        other : AudioRegion or int
            An `AudioRegion` with the same audio parameters as this region, or
            `0` to enable concatenating multiple regions using `sum`.

        Returns
        -------
        AudioRegion
            A new `AudioRegion` representing the concatenation result.
        """
        if other == 0:
            return self
        return other.add(self)

    def __mul__(self, n):
        if not isinstance(n, int):
            err_msg = "Can't multiply AudioRegion by a non-int of type '{}'"
            raise TypeError(err_msg.format(type(n)))
        data = self.data * n
        return AudioRegion(data, self.sr, self.sw, self.ch)

    def __rmul__(self, n):
        return self * n

    def __truediv__(self, n):
        if not isinstance(n, int) or n <= 0:
            raise TypeError("AudioRegion can only be divided by a positive int")
        samples_per_sub_region, rest = divmod(len(self), n)
        onset = 0
        sub_regions = []
        while onset < len(self):
            offset = 0
            if rest > 0:
                offset = 1
                rest -= 1
            offset += onset + samples_per_sub_region
            sub_regions.append(self[onset:offset])
            onset = offset
        return sub_regions

    def __eq__(self, other):
        if other is self:
            return True
        if not isinstance(other, AudioRegion):
            return False
        return (
            (self.data == other.data)
            and (self.sr == other.sr)
            and (self.sw == other.sw)
            and (self.ch == other.ch)
        )

    def __getitem__(self, index):
        err_msg = "Slicing AudioRegion by samples requires indices of type "
        err_msg += "'int' without a step (e.g. region.sec[1600:3200])"
        start_sample, stop_sample = _check_convert_index(index, (int), err_msg)

        bytes_per_sample = self.sample_width * self.channels
        len_samples = len(self.data) // bytes_per_sample

        if start_sample < 0:
            start_sample = max(start_sample + len_samples, 0)
        onset = start_sample * bytes_per_sample

        if stop_sample is not None:
            if stop_sample < 0:
                stop_sample = max(stop_sample + len_samples, 0)
            offset = index.stop * bytes_per_sample
        else:
            offset = None

        data = self.data[onset:offset]
        return AudioRegion(data, self.sr, self.sw, self.ch)


class StreamTokenizer:
    """
    Class for stream tokenizers, implementing a 4-state automaton scheme
    to extract relevant sub-sequences from a data stream in real-time.

    Parameters
    ----------
    validator : callable or :class:`DataValidator` (must implement `is_valid`).
        Called with each data frame read from the source. Should take a
        single argument and return True or False to indicate valid and
        invalid frames, respectively.

    min_length : int
        Minimum number of frames in a valid token, including any tolerated
        non-valid frames within the token.

    max_length : int
        Maximum number of frames in a valid token, including all tolerated
        non-valid frames within the token.

    max_continuous_silence : int
        Maximum number of consecutive non-valid frames within a token. Each
        silent region may contain up to `max_continuous_silence` frames.

    init_min : int, default=0
        Minimum number of consecutive valid frames required before
        tolerating any non-valid frames. Helps discard non-valid tokens
        early if needed.

    init_max_silence : int, default=0
        Maximum number of tolerated consecutive non-valid frames before
        reaching `init_min`. Used if `init_min` is specified.

    mode : int
        Defines the tokenizer behavior with the following options:

        - `StreamTokenizer.NORMAL` (0, default): Do not drop trailing silence
          and allow tokens shorter than `min_length` if they immediately follow
          a delivered token.

        - `StreamTokenizer.STRICT_MIN_LENGTH` (2): If a token `i` is
          delivered at `max_length`, any adjacent token `i+1` must meet
          `min_length`.

        - `StreamTokenizer.DROP_TRAILING_SILENCE` (4): Drop all trailing
          non-valid frames from a token unless the token is truncated
          (e.g., at `max_length`).

        - `StreamTokenizer.STRICT_MIN_LENGTH | StreamTokenizer.DROP_TRAILING_SILENCE`:
          Apply both `STRICT_MIN_LENGTH` and `DROP_TRAILING_SILENCE`.

    Examples
    --------
    In the following, without `STRICT_MIN_LENGTH`, the 'BB' token is
    accepted even though it is shorter than `min_length` (3) because it
    immediately follows the last delivered token:

    >>> from auditok.core import StreamTokenizer
    >>> from auditok.util import StringDataSource, DataValidator

    >>> class UpperCaseChecker(DataValidator):
    >>>     def is_valid(self, frame):
                return frame.isupper()

    >>> dsource = StringDataSource("aaaAAAABBbbb")
    >>> tokenizer = StreamTokenizer(
    >>>     validator=UpperCaseChecker(),
    >>>     min_length=3,
    >>>     max_length=4,
    >>>     max_continuous_silence=0
    >>> )
    >>> tokenizer.tokenize(dsource)
    [(['A', 'A', 'A', 'A'], 3, 6), (['B', 'B'], 7, 8)]

    Using `STRICT_MIN_LENGTH` mode rejects the 'BB' token:

    >>> tokenizer = StreamTokenizer(
    >>>     validator=UpperCaseChecker(),
    >>>     min_length=3,
    >>>     max_length=4,
    >>>     max_continuous_silence=0,
    >>>     mode=StreamTokenizer.STRICT_MIN_LENGTH
    >>> )
    >>> tokenizer.tokenize(dsource)
    [(['A', 'A', 'A', 'A'], 3, 6)]

    With `DROP_TRAILING_SILENCE`, trailing silence is removed if not truncated:

    >>> tokenizer = StreamTokenizer(
    >>>     validator=UpperCaseChecker(),
    >>>     min_length=3,
    >>>     max_length=6,
    >>>     max_continuous_silence=3,
    >>>     mode=StreamTokenizer.DROP_TRAILING_SILENCE
    >>> )
    >>> dsource = StringDataSource("aaaAAAaaaBBbbbb")
    >>> tokenizer.tokenize(dsource)
    [(['A', 'A', 'A', 'a', 'a', 'a'], 3, 8), (['B', 'B'], 9, 10)]

    Without `DROP_TRAILING_SILENCE`, the output includes trailing frames:

    .. code:: python

        [
            (['A', 'A', 'A', 'a', 'a', 'a'], 3, 8),
            (['B', 'B', 'b', 'b', 'b'], 9, 13)
        ]
    """

    SILENCE = 0
    POSSIBLE_SILENCE = 1
    POSSIBLE_NOISE = 2
    NOISE = 3
    NORMAL = 0
    STRICT_MIN_LENGTH = 2
    DROP_TRAILING_SILENCE = 4

    def __init__(
        self,
        validator,
        min_length,
        max_length,
        max_continuous_silence,
        init_min=0,
        init_max_silence=0,
        mode=0,
    ):
        if callable(validator):
            self._is_valid = validator
        elif isinstance(validator, DataValidator):
            self._is_valid = validator.is_valid
        else:
            raise TypeError(
                "'validator' must be a callable or an instance of "
                "DataValidator"
            )

        if max_length <= 0:
            raise ValueError(
                "'max_length' must be > 0 (value={0})".format(max_length)
            )

        if min_length <= 0 or min_length > max_length:
            err_msg = "'min_length' must be > 0 and <= 'max_length' (value={0})"
            raise ValueError(err_msg.format(min_length))

        if max_continuous_silence >= max_length:
            err_msg = "'max_continuous_silence' must be < 'max_length' "
            err_msg += "(value={0})"
            raise ValueError(err_msg.format(max_continuous_silence))

        if init_min >= max_length:
            raise ValueError(
                "'init_min' must be < 'max_length' (value={0})".format(
                    max_continuous_silence
                )
            )

        self.validator = validator
        self.min_length = min_length
        self.max_length = max_length
        self.max_continuous_silence = max_continuous_silence
        self.init_min = init_min
        self.init_max_silent = init_max_silence
        self._set_mode(mode)
        self._deliver = None
        self._tokens = None
        self._state = None
        self._data = None
        self._contiguous_token = False
        self._init_count = 0
        self._silence_length = 0
        self._start_frame = 0
        self._current_frame = 0

    def _set_mode(self, mode):
        strict_min_and_drop_trailing = StreamTokenizer.STRICT_MIN_LENGTH
        strict_min_and_drop_trailing |= StreamTokenizer.DROP_TRAILING_SILENCE
        if mode not in [
            StreamTokenizer.NORMAL,
            StreamTokenizer.STRICT_MIN_LENGTH,
            StreamTokenizer.DROP_TRAILING_SILENCE,
            strict_min_and_drop_trailing,
        ]:
            raise ValueError("Wrong value for mode")
        self._mode = mode
        self._strict_min_length = (mode & self.STRICT_MIN_LENGTH) != 0
        self._drop_trailing_silence = (mode & self.DROP_TRAILING_SILENCE) != 0

    def _reinitialize(self):
        self._contiguous_token = False
        self._data = []
        self._tokens = []
        self._state = self.SILENCE
        self._current_frame = -1
        self._deliver = self._append_token

    def tokenize(self, data_source, callback=None, generator=False):
        """
        Read data from `data_source` one frame at a time and process each frame
        to detect sequences that form valid tokens.

        Parameters
        ----------
        data_source : DataSource
            An instance of the :class:`DataSource` class that implements a `read`
            method. `read` should return a slice of the signal (a frame of any
            type that can be processed by the validator) or None when there is no
            more data in the source.

        callback : callable, optional
            A function that takes three arguments. If provided, `callback` is
            called each time a valid token is detected.

        generator : bool, optional, default=False
            If True, the method yields tokens as they are detected, rather than
            returning a list. If False, a list of tokens is returned.

        Returns
        -------
        list of tuples or generator
            A list of tokens if `generator` is False, or a generator yielding
            tokens if `generator` is True. Each token is a tuple with the
            following structure:

            .. code:: python

                (data, start, end)

            where `data` is a list of frames in the token, `start` is the index
            of the first frame in the original data, and `end` is the index of
            the last frame.
        """

        token_gen = self._iter_tokens(data_source)
        if callback:
            for token in token_gen:
                callback(*token)
            return
        if generator:
            return token_gen
        return list(token_gen)

    def _iter_tokens(self, data_source):
        self._reinitialize()
        while True:
            frame = data_source.read()
            self._current_frame += 1
            if frame is None:
                token = self._post_process()
                if token is not None:
                    yield token
                break
            token = self._process(frame)
            if token is not None:
                yield token

    def _process(self, frame):  # noqa: C901

        frame_is_valid = self._is_valid(frame)

        if self._state == self.SILENCE:

            if frame_is_valid:
                # seems we got a valid frame after a silence
                self._init_count = 1
                self._silence_length = 0
                self._start_frame = self._current_frame
                self._data.append(frame)

                if self._init_count >= self.init_min:
                    self._state = self.NOISE
                    if len(self._data) >= self.max_length:
                        return self._process_end_of_detection(True)
                else:
                    self._state = self.POSSIBLE_NOISE

        elif self._state == self.POSSIBLE_NOISE:

            if frame_is_valid:
                self._silence_length = 0
                self._init_count += 1
                self._data.append(frame)
                if self._init_count >= self.init_min:
                    self._state = self.NOISE
                    if len(self._data) >= self.max_length:
                        return self._process_end_of_detection(True)

            else:
                self._silence_length += 1
                if (
                    self._silence_length > self.init_max_silent
                    or len(self._data) + 1 >= self.max_length
                ):
                    # either init_max_silent or max_length is reached
                    # before _init_count, back to silence
                    self._data = []
                    self._state = self.SILENCE
                else:
                    self._data.append(frame)

        elif self._state == self.NOISE:

            if frame_is_valid:
                self._data.append(frame)
                if len(self._data) >= self.max_length:
                    return self._process_end_of_detection(True)

            elif self.max_continuous_silence <= 0:
                # max token reached at this frame will _deliver if
                # _contiguous_token and not _strict_min_length
                self._state = self.SILENCE
                return self._process_end_of_detection()
            else:
                # this is the first silent frame following a valid one
                # and it is tolerated
                self._silence_length = 1
                self._data.append(frame)
                self._state = self.POSSIBLE_SILENCE
                if len(self._data) == self.max_length:
                    return self._process_end_of_detection(True)
                    # don't reset _silence_length because we still
                    # need to know the total number of silent frames

        elif self._state == self.POSSIBLE_SILENCE:

            if frame_is_valid:
                self._data.append(frame)
                self._silence_length = 0
                self._state = self.NOISE
                if len(self._data) >= self.max_length:
                    return self._process_end_of_detection(True)

            else:
                if self._silence_length >= self.max_continuous_silence:
                    self._state = self.SILENCE
                    if self._silence_length < len(self._data):
                        # _deliver only gathered frames aren't all silent
                        return self._process_end_of_detection()
                    self._data = []
                    self._silence_length = 0
                else:
                    self._data.append(frame)
                    self._silence_length += 1
                    if len(self._data) >= self.max_length:
                        return self._process_end_of_detection(True)
                        # don't reset _silence_length because we still
                        # need to know the total number of silent frames

    def _post_process(self):
        if self._state == self.NOISE or self._state == self.POSSIBLE_SILENCE:
            if len(self._data) > 0 and len(self._data) > self._silence_length:
                return self._process_end_of_detection()

    def _process_end_of_detection(self, truncated=False):

        if (
            not truncated
            and self._drop_trailing_silence
            and self._silence_length > 0
        ):
            # happens if max_continuous_silence is reached
            # or max_length is reached at a silent frame
            self._data = self._data[0 : -self._silence_length]

        if (len(self._data) >= self.min_length) or (
            len(self._data) > 0
            and not self._strict_min_length
            and self._contiguous_token
        ):

            start_frame = self._start_frame
            end_frame = self._start_frame + len(self._data) - 1
            data = self._data
            self._data = []
            token = (data, start_frame, end_frame)

            if truncated:
                # next token (if any) will start at _current_frame + 1
                self._start_frame = self._current_frame + 1
                # remember that it is contiguous with the just delivered one
                self._contiguous_token = True
            else:
                self._contiguous_token = False
            return token
        else:
            self._contiguous_token = False

        self._data = []

    def _append_token(self, data, start, end):
        self._tokens.append((data, start, end))



import filecmp
import os
import wave
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from unittest.mock import Mock, patch

import numpy as np
import pytest
from test_AudioSource import PURE_TONE_DICT, _sample_generator

import auditok
from auditok.io import (
    AudioIOError,
    AudioParameterError,
    BufferAudioSource,
    RawAudioSource,
    StdinAudioSource,
    WaveAudioSource,
    _get_audio_parameters,
    _guess_audio_format,
    _load_raw,
    _load_wave,
    _load_with_pydub,
    _save_raw,
    _save_wave,
    _save_with_pydub,
    check_audio_data,
    from_file,
    get_audio_source,
    to_file,
)
from auditok.signal import SAMPLE_WIDTH_TO_DTYPE

AUDIO_PARAMS = {"sampling_rate": 16000, "sample_width": 2, "channels": 1}
AUDIO_PARAMS_SHORT = {"sr": 16000, "sw": 2, "ch": 1}


@pytest.mark.parametrize(
    "data, sample_width, channels, valid",
    [
        (b"\0" * 113, 1, 1, True),  # valid_mono
        (b"\0" * 160, 1, 2, True),  # valid_stereo
        (b"\0" * 113, 2, 1, False),  # invalid_mono_sw_2
        (b"\0" * 113, 1, 2, False),  # invalid_stereo_sw_1
        (b"\0" * 158, 2, 2, False),  # invalid_stereo_sw_2
    ],
    ids=[
        "valid_mono",
        "valid_stereo",
        "invalid_mono_sw_2",
        "invalid_stereo_sw_1",
        "invalid_stereo_sw_2",
    ],
)
def test_check_audio_data(data, sample_width, channels, valid):
    if not valid:
        with pytest.raises(AudioParameterError):
            check_audio_data(data, sample_width, channels)
    else:
        assert check_audio_data(data, sample_width, channels) is None


@pytest.mark.parametrize(
    "filename, audio_format, expected",
    [
        ("filename.wav", "wav", "wav"),  # extension_and_format_same
        ("filename.mp3", "wav", "wav"),  # extension_and_format_different
        ("filename.wav", None, "wav"),  # extension_no_format
        ("filename", "wav", "wav"),  # format_no_extension
        ("filename", None, None),  # no_format_no_extension
        ("filename", "wave", "wav"),  # wave_as_wav
        ("filename.wave", None, "wav"),  # wave_as_wav_extension
    ],
    ids=[
        "extension_and_format_same",
        "extension_and_format_different",
        "extension_no_format",
        "format_no_extension",
        "no_format_no_extension",
        "wave_as_wav",
        "wave_as_wav_extension",
    ],
)
def test_guess_audio_format(filename, audio_format, expected):
    result = _guess_audio_format(filename, audio_format)
    assert result == expected

    result = _guess_audio_format(Path(filename), audio_format)
    assert result == expected


def test_get_audio_parameters_short_params():
    expected = (8000, 2, 1)
    params = dict(zip(("sr", "sw", "ch"), expected))
    result = _get_audio_parameters(params)
    assert result == expected


def test_get_audio_parameters_long_params():
    expected = (8000, 2, 1)
    params = dict(zip(("sampling_rate", "sample_width", "channels"), expected))
    result = _get_audio_parameters(params)
    assert result == expected


def test_get_audio_parameters_long_params_shadow_short_ones():
    expected = (8000, 2, 1)
    params = dict(
        zip(
            ("sampling_rate", "sample_width", "channels"),
            expected,
        )
    )
    params.update(
        dict(
            zip(
                ("sr", "sw", "ch"),
                "xxx",
            )
        )
    )
    result = _get_audio_parameters(params)
    assert result == expected


@pytest.mark.parametrize(
    "missing_param",
    [
        "sampling_rate",  # missing_sampling_rate
        "sample_width",  # missing_sample_width
        "channels",  # missing_channels
    ],
    ids=["missing_sampling_rate", "missing_sample_width", "missing_channels"],
)
def test_get_audio_parameters_missing_parameter(missing_param):
    params = AUDIO_PARAMS.copy()
    del params[missing_param]
    with pytest.raises(AudioParameterError):
        _get_audio_parameters(params)


@pytest.mark.parametrize(
    "missing_param",
    [
        "sr",  # missing_sampling_rate
        "sw",  # missing_sample_width
        "ch",  # missing_channels
    ],
    ids=["missing_sampling_rate", "missing_sample_width", "missing_channels"],
)
def test_get_audio_parameters_missing_parameter_short(missing_param):
    params = AUDIO_PARAMS_SHORT.copy()
    del params[missing_param]
    with pytest.raises(AudioParameterError):
        _get_audio_parameters(params)


@pytest.mark.parametrize(
    "values",
    [
        ("x", 2, 1),  # str_sampling_rate
        (-8000, 2, 1),  # negative_sampling_rate
        (8000, "x", 1),  # str_sample_width
        (8000, -2, 1),  # negative_sample_width
        (8000, 2, "x"),  # str_channels
        (8000, 2, -1),  # negative_channels
    ],
    ids=[
        "str_sampling_rate",
        "negative_sampling_rate",
        "str_sample_width",
        "negative_sample_width",
        "str_channels",
        "negative_channels",
    ],
)
def test_get_audio_parameters_invalid(values):
    params = dict(
        zip(
            ("sampling_rate", "sample_width", "channels"),
            values,
        )
    )
    with pytest.raises(AudioParameterError):
        _get_audio_parameters(params)


@pytest.mark.parametrize(
    "filename, audio_format, funtion_name, kwargs",
    [
        (
            "audio",
            "raw",
            "_load_raw",
            AUDIO_PARAMS_SHORT,
        ),  # raw_with_audio_format
        (
            "audio.raw",
            None,
            "_load_raw",
            AUDIO_PARAMS_SHORT,
        ),  # raw_with_extension
        ("audio", "wave", "_load_wave", None),  # wave_with_audio_format
        ("audio", "wave", "_load_wave", None),  # wav_with_audio_format
        ("audio.wav", None, "_load_wave", None),  # wav_with_extension
        (
            "audio.dat",
            "wav",
            "_load_wave",
            None,
        ),  # format_and_extension_both_given_a
        (
            "audio.raw",
            "wave",
            "_load_wave",
            None,
        ),  # format_and_extension_both_given_b
        ("audio", None, "_load_with_pydub", None),  # no_format_nor_extension
        ("audio.ogg", None, "_load_with_pydub", None),  # other_formats_ogg
        ("audio", "webm", "_load_with_pydub", None),  # other_formats_webm
    ],
    ids=[
        "raw_with_audio_format",
        "raw_with_extension",
        "wave_with_audio_format",
        "wav_with_audio_format",
        "wav_with_extension",
        "format_and_extension_both_given_a",
        "format_and_extension_both_given_b",
        "no_format_nor_extension",
        "other_formats_ogg",
        "other_formats_webm",
    ],
)
def test_from_file(filename, audio_format, funtion_name, kwargs):
    funtion_name = "auditok.io." + funtion_name
    if kwargs is None:
        kwargs = {}
    with patch(funtion_name) as patch_function:
        from_file(filename, audio_format, **kwargs)
    assert patch_function.called


@pytest.mark.parametrize(
    "large_file, cls, size, use_pathlib",
    [
        (False, BufferAudioSource, -1, False),  # large_file_false_negative_size
        (False, BufferAudioSource, None, False),  # large_file_false_None_size
        (
            False,
            BufferAudioSource,
            None,
            True,
        ),  # large_file_false_None_size_Path
        (True, RawAudioSource, -1, False),  # large_file_true_negative_size
        (True, RawAudioSource, None, False),  # large_file_true_None_size
        (True, RawAudioSource, -1, True),  # large_file_true_negative_size_Path
    ],
    ids=[
        "large_file_false_negative_size",
        "large_file_false_None_size",
        "large_file_false_None_size_Path",
        "large_file_true_negative_size",
        "large_file_true_None_size",
        "large_file_true_negative_size_Path",
    ],
)
def test_from_file_raw_read_all(large_file, cls, size, use_pathlib):
    filename = Path("tests/data/test_16KHZ_mono_400Hz.raw")
    if use_pathlib:
        filename = Path(filename)
    audio_source = from_file(
        filename,
        large_file=large_file,
        sampling_rate=16000,
        sample_width=2,
        channels=1,
    )
    assert isinstance(audio_source, cls)

    with open(filename, "rb") as fp:
        expected = fp.read()
    audio_source.open()
    data = audio_source.read(size)
    audio_source.close()
    assert data == expected


@pytest.mark.parametrize(
    "large_file, cls, size, use_pathlib",
    [
        (False, BufferAudioSource, -1, False),  # large_file_false_negative_size
        (False, BufferAudioSource, None, False),  # large_file_false_None_size
        (
            False,
            BufferAudioSource,
            None,
            True,
        ),  # large_file_false_None_size_Path
        (True, WaveAudioSource, -1, False),  # large_file_true_negative_size
        (True, WaveAudioSource, None, False),  # large_file_true_None_size
        (True, WaveAudioSource, -1, True),  # large_file_true_negative_size_Path
    ],
    ids=[
        "large_file_false_negative_size",
        "large_file_false_None_size",
        "large_file_false_None_size_Path",
        "large_file_true_negative_size",
        "large_file_true_None_size",
        "large_file_true_negative_size_Path",
    ],
)
def test_from_file_wave_read_all(large_file, cls, size, use_pathlib):
    filename = "tests/data/test_16KHZ_mono_400Hz.wav"
    if use_pathlib:
        filename = Path(filename)
    audio_source = from_file(
        filename,
        large_file=large_file,
        sampling_rate=16000,
        sample_width=2,
        channels=1,
    )
    assert isinstance(audio_source, cls)

    with wave.open(str(filename)) as fp:
        expected = fp.readframes(-1)
    audio_source.open()
    data = audio_source.read(size)
    audio_source.close()
    assert data == expected


def test_from_file_large_file_compressed():
    filename = "tests/data/test_16KHZ_mono_400Hz.ogg"
    with pytest.raises(AudioIOError):
        from_file(filename, large_file=True)


@pytest.mark.parametrize(
    "missing_param",
    [
        "sr",  # missing_sampling_rate
        "sw",  # missing_sample_width
        "ch",  # missing_channels
    ],
    ids=["missing_sampling_rate", "missing_sample_width", "missing_channels"],
)
def test_from_file_missing_audio_param(missing_param):
    params = AUDIO_PARAMS_SHORT.copy()
    del params[missing_param]
    with pytest.raises(AudioParameterError):
        from_file("audio", audio_format="raw", **params)


def test_from_file_no_pydub():
    with patch("auditok.io._WITH_PYDUB", False):
        with pytest.raises(AudioIOError):
            from_file("audio", "mp3")


@pytest.mark.parametrize(
    "audio_format, function",
    [
        ("ogg", "from_ogg"),  # ogg_first_channel
        ("ogg", "from_ogg"),  # ogg_second_channel
        ("ogg", "from_ogg"),  # ogg_mix
        ("ogg", "from_ogg"),  # ogg_default
        ("mp3", "from_mp3"),  # mp3_left_channel
        ("mp3", "from_mp3"),  # mp3_right_channel
        ("flac", "from_file"),  # flac_first_channel
        ("flac", "from_file"),  # flac_second_channel
        ("flv", "from_flv"),  # flv_left_channel
        ("webm", "from_file"),  # webm_right_channel
    ],
    ids=[
        "ogg_first_channel",
        "ogg_second_channel",
        "ogg_mix",
        "ogg_default",
        "mp3_left_channel",
        "mp3_right_channel",
        "flac_first_channel",
        "flac_second_channel",
        "flv_left_channel",
        "webm_right_channel",
    ],
)
@patch("auditok.io._WITH_PYDUB", True)
@patch("auditok.io.BufferAudioSource")
def test_from_file_multichannel_audio_compressed(
    mock_buffer_audio_source, audio_format, function
):
    filename = "audio.{}".format(audio_format)
    segment_mock = Mock()
    segment_mock.sample_width = 2
    segment_mock.channels = 2
    segment_mock._data = b"abcd"
    with patch("auditok.io.AudioSegment.{}".format(function)) as open_func:
        open_func.return_value = segment_mock
        from_file(filename)
        assert open_func.called


@pytest.mark.parametrize(
    "file_id, frequencies, large_file",
    [
        ("mono_400", (400,), False),  # mono
        ("3channel_400-800-1600", (400, 800, 1600), False),  # three_channel
        ("mono_400", (400,), True),  # mono_large_file
        (
            "3channel_400-800-1600",
            (400, 800, 1600),
            True,
        ),  # three_channel_large_file
    ],
    ids=[
        "mono",
        "three_channel",
        "mono_large_file",
        "three_channel_large_file",
    ],
)
def test_load_raw(file_id, frequencies, large_file):
    filename = "tests/data/test_16KHZ_{}Hz.raw".format(file_id)
    audio_source = _load_raw(
        filename, 16000, 2, len(frequencies), large_file=large_file
    )
    audio_source.open()
    data = audio_source.read(-1)
    audio_source.close()
    expected_class = RawAudioSource if large_file else BufferAudioSource
    assert isinstance(audio_source, expected_class)
    assert audio_source.sampling_rate == 16000
    assert audio_source.sample_width == 2
    assert audio_source.channels == len(frequencies)
    mono_channels = [PURE_TONE_DICT[freq] for freq in frequencies]
    dtype = SAMPLE_WIDTH_TO_DTYPE[audio_source.sample_width]
    expected = np.fromiter(
        _sample_generator(*mono_channels), dtype=dtype
    ).tobytes()
    assert data == expected


def test_load_raw_missing_audio_param():
    with pytest.raises(AudioParameterError):
        _load_raw("audio", sampling_rate=None, sample_width=1, channels=1)

    with pytest.raises(AudioParameterError):
        _load_raw("audio", sampling_rate=16000, sample_width=None, channels=1)

    with pytest.raises(AudioParameterError):
        _load_raw("audio", sampling_rate=16000, sample_width=1, channels=None)


@pytest.mark.parametrize(
    "file_id, frequencies, large_file",
    [
        ("mono_400", (400,), False),  # mono
        ("3channel_400-800-1600", (400, 800, 1600), False),  # three_channel
        ("mono_400", (400,), True),  # mono_large_file
        (
            "3channel_400-800-1600",
            (400, 800, 1600),
            True,
        ),  # three_channel_large_file
    ],
    ids=[
        "mono",
        "three_channel",
        "mono_large_file",
        "three_channel_large_file",
    ],
)
def test_load_wave(file_id, frequencies, large_file):
    filename = "tests/data/test_16KHZ_{}Hz.wav".format(file_id)
    audio_source = _load_wave(filename, large_file=large_file)
    audio_source.open()
    data = audio_source.read(-1)
    audio_source.close()
    expected_class = WaveAudioSource if large_file else BufferAudioSource
    assert isinstance(audio_source, expected_class)
    assert audio_source.sampling_rate == 16000
    assert audio_source.sample_width == 2
    assert audio_source.channels == len(frequencies)
    mono_channels = [PURE_TONE_DICT[freq] for freq in frequencies]
    dtype = SAMPLE_WIDTH_TO_DTYPE[audio_source.sample_width]
    expected = np.fromiter(
        _sample_generator(*mono_channels), dtype=dtype
    ).tobytes()
    assert data == expected


@pytest.mark.parametrize(
    "audio_format, channels, function",
    [
        ("ogg", 2, "from_ogg"),  # ogg_default_first_channel
        ("ogg", 1, "from_ogg"),  # ogg_first_channel
        ("ogg", 2, "from_ogg"),  # ogg_second_channel
        ("ogg", 3, "from_ogg"),  # ogg_mix_channels
        ("mp3", 1, "from_mp3"),  # mp3_left_channel
        ("mp3", 2, "from_mp3"),  # mp3_right_channel
        ("mp3", 3, "from_mp3"),  # mp3_mix_channels
        ("flac", 2, "from_file"),  # flac_first_channel
        ("flac", 2, "from_file"),  # flac_second_channel
        ("flv", 1, "from_flv"),  # flv_left_channel
        ("webm", 2, "from_file"),  # webm_right_channel
        ("webm", 4, "from_file"),  # webm_mix_channels
    ],
    ids=[
        "ogg_default_first_channel",
        "ogg_first_channel",
        "ogg_second_channel",
        "ogg_mix_channels",
        "mp3_left_channel",
        "mp3_right_channel",
        "mp3_mix_channels",
        "flac_first_channel",
        "flac_second_channel",
        "flv_left_channel",
        "webm_right_channel",
        "webm_mix_channels",
    ],
)
@patch("auditok.io._WITH_PYDUB", True)
@patch("auditok.io.BufferAudioSource")
def test_load_with_pydub(
    mock_buffer_audio_source, audio_format, channels, function
):
    filename = "audio.{}".format(audio_format)
    segment_mock = Mock()
    segment_mock.sample_width = 2
    segment_mock.channels = channels
    segment_mock._data = b"abcdefgh"
    with patch("auditok.io.AudioSegment.{}".format(function)) as open_func:
        open_func.return_value = segment_mock
        _load_with_pydub(filename, audio_format)
        assert open_func.called


@pytest.mark.parametrize(
    "filename, frequencies, use_pathlib",
    [
        ("mono_400Hz.raw", (400,), False),  # mono
        ("mono_400Hz.raw", (400,), True),  # mono_pathlib
        (
            "3channel_400-800-1600Hz.raw",
            (400, 800, 1600),
            False,
        ),  # three_channel
    ],
    ids=["mono", "three_channel", "use_pathlib"],
)
def test_save_raw(filename, frequencies, use_pathlib):
    filename = "tests/data/test_16KHZ_{}".format(filename)
    if use_pathlib:
        filename = Path(filename)
    sample_width = 2
    dtype = SAMPLE_WIDTH_TO_DTYPE[sample_width]
    mono_channels = [PURE_TONE_DICT[freq] for freq in frequencies]
    data = np.fromiter(_sample_generator(*mono_channels), dtype=dtype).tobytes()
    tmpfile = NamedTemporaryFile()
    _save_raw(data, tmpfile.name)
    assert filecmp.cmp(tmpfile.name, filename, shallow=False)


@pytest.mark.parametrize(
    "filename, frequencies, use_pathlib",
    [
        ("mono_400Hz.wav", (400,), False),  # mono
        ("mono_400Hz.wav", (400,), True),  # mono_pathlib
        (
            "3channel_400-800-1600Hz.wav",
            (400, 800, 1600),
            False,
        ),  # three_channel
    ],
    ids=["mono", "mono_pathlib", "three_channel"],
)
def test_save_wave(filename, frequencies, use_pathlib):
    filename = "tests/data/test_16KHZ_{}".format(filename)
    if use_pathlib:
        filename = str(filename)
    sampling_rate = 16000
    sample_width = 2
    channels = len(frequencies)
    mono_channels = [PURE_TONE_DICT[freq] for freq in frequencies]
    dtype = SAMPLE_WIDTH_TO_DTYPE[sample_width]
    data = np.fromiter(_sample_generator(*mono_channels), dtype=dtype).tobytes()
    tmpfile = NamedTemporaryFile()
    _save_wave(data, tmpfile.name, sampling_rate, sample_width, channels)
    assert filecmp.cmp(tmpfile.name, filename, shallow=False)


@pytest.mark.parametrize(
    "missing_param",
    [
        "sr",  # missing_sampling_rate
        "sw",  # missing_sample_width
        "ch",  # missing_channels
    ],
    ids=["missing_sampling_rate", "missing_sample_width", "missing_channels"],
)
def test_save_wave_missing_audio_param(missing_param):
    with pytest.raises(AudioParameterError):
        _save_wave(
            b"\0\0", "audio", sampling_rate=None, sample_width=1, channels=1
        )

    with pytest.raises(AudioParameterError):
        _save_wave(
            b"\0\0", "audio", sampling_rate=16000, sample_width=None, channels=1
        )

    with pytest.raises(AudioParameterError):
        _save_wave(
            b"\0\0", "audio", sampling_rate=16000, sample_width=1, channels=None
        )


def test_save_with_pydub():
    with patch("auditok.io.AudioSegment.export") as export:
        tmpdir = TemporaryDirectory()
        filename = os.path.join(tmpdir.name, "audio.ogg")
        _save_with_pydub(b"\0\0", filename, "ogg", 16000, 2, 1)
        assert export.called
        tmpdir.cleanup()


@pytest.mark.parametrize(
    "filename, audio_format",
    [
        ("audio", "raw"),  # raw_with_audio_format
        ("audio.raw", None),  # raw_with_extension
        ("audio.mp3", "raw"),  # raw_with_audio_format_and_extension
        ("audio", None),  # raw_no_audio_format_nor_extension
    ],
    ids=[
        "raw_with_audio_format",
        "raw_with_extension",
        "raw_with_audio_format_and_extension",
        "raw_no_audio_format_nor_extension",
    ],
)
def test_to_file_raw(filename, audio_format):
    exp_filename = "tests/data/test_16KHZ_mono_400Hz.raw"
    tmpdir = TemporaryDirectory()
    filename = os.path.join(tmpdir.name, filename)
    data = PURE_TONE_DICT[400].tobytes()
    to_file(data, filename, audio_format=audio_format)
    assert filecmp.cmp(filename, exp_filename, shallow=False)
    tmpdir.cleanup()


@pytest.mark.parametrize(
    "filename, audio_format",
    [
        ("audio", "wav"),  # wav_with_audio_format
        ("audio.wav", None),  # wav_with_extension
        ("audio.mp3", "wav"),  # wav_with_audio_format_and_extension
        ("audio", "wave"),  # wave_with_audio_format
        ("audio.wave", None),  # wave_with_extension
        ("audio.mp3", "wave"),  # wave_with_audio_format_and_extension
    ],
    ids=[
        "wav_with_audio_format",
        "wav_with_extension",
        "wav_with_audio_format_and_extension",
        "wave_with_audio_format",
        "wave_with_extension",
        "wave_with_audio_format_and_extension",
    ],
)
def test_to_file_wave(filename, audio_format):
    exp_filename = "tests/data/test_16KHZ_mono_400Hz.wav"
    tmpdir = TemporaryDirectory()
    filename = os.path.join(tmpdir.name, filename)
    data = PURE_TONE_DICT[400].tobytes()
    to_file(
        data,
        filename,
        audio_format=audio_format,
        sampling_rate=16000,
        sample_width=2,
        channels=1,
    )
    assert filecmp.cmp(filename, exp_filename, shallow=False)
    tmpdir.cleanup()


@pytest.mark.parametrize(
    "missing_param",
    [
        "sr",  # missing_sampling_rate
        "sw",  # missing_sample_width
        "ch",  # missing_channels
    ],
    ids=["missing_sampling_rate", "missing_sample_width", "missing_channels"],
)
def test_to_file_missing_audio_param(missing_param):
    params = AUDIO_PARAMS_SHORT.copy()
    del params[missing_param]
    with pytest.raises(AudioParameterError):
        to_file(b"\0\0", "audio", audio_format="wav", **params)
    with pytest.raises(AudioParameterError):
        to_file(b"\0\0", "audio", audio_format="mp3", **params)


def test_to_file_no_pydub():
    with patch("auditok.io._WITH_PYDUB", False):
        with pytest.raises(AudioIOError):
            to_file("audio", b"", "mp3")


@pytest.mark.parametrize(
    "filename, audio_format",
    [
        ("audio.ogg", None),  # ogg_with_extension
        ("audio", "ogg"),  # ogg_with_audio_format
        ("audio.wav", "ogg"),  # ogg_format_with_wrong_extension
    ],
    ids=[
        "ogg_with_extension",
        "ogg_with_audio_format",
        "ogg_format_with_wrong_extension",
    ],
)
@patch("auditok.io._WITH_PYDUB", True)
def test_to_file_compressed(filename, audio_format):
    with patch("auditok.io.AudioSegment.export") as export:
        tmpdir = TemporaryDirectory()
        filename = os.path.join(tmpdir.name, filename)
        to_file(b"\0\0", filename, audio_format, **AUDIO_PARAMS_SHORT)
        assert export.called
        tmpdir.cleanup()


@pytest.mark.parametrize(
    "input, expected_type, extra_args",
    [
        (
            "tests/data/test_16KHZ_mono_400Hz.wav",
            BufferAudioSource,
            None,
        ),  # string_wave
        (
            "tests/data/test_16KHZ_mono_400Hz.wav",
            WaveAudioSource,
            {"large_file": True},
        ),  # string_wave_large_file
        ("-", StdinAudioSource, None),  # stdin
        (
            "tests/data/test_16KHZ_mono_400Hz.raw",
            BufferAudioSource,
            None,
        ),  # string_raw
        (
            "tests/data/test_16KHZ_mono_400Hz.raw",
            RawAudioSource,
            {"large_file": True},
        ),  # string_raw_large_file
        (b"0" * 8000, BufferAudioSource, None),  # bytes_
    ],
    ids=[
        "string_wave",
        "string_wave_large_file",
        "stdin",
        "string_raw",
        "string_raw_large_file",
        "bytes_",
    ],
)
def test_get_audio_source(input, expected_type, extra_args):
    kwargs = {"sampling_rate": 16000, "sample_width": 2, "channels": 1}
    if extra_args is not None:
        kwargs.update(extra_args)
    audio_source = get_audio_source(input, **kwargs)
    assert isinstance(audio_source, expected_type)
    assert audio_source.sampling_rate == 16000, (
        "Unexpected sampling rate: audio_source.sampling_rate = "
        + f"{audio_source.sampling_rate} instead of 16000"
    )
    assert audio_source.sr == 16000, (
        "Unexpected sampling rate: audio_source.sr = "
        + f"{audio_source.sr} instead of 16000"
    )
    assert audio_source.sample_width == 2, (
        "Unexpected sample width: audio_source.sample_width = "
        + f"{audio_source.sample_width} instead of 2"
    )
    assert audio_source.sw == 2, (
        "Unexpected sample width: audio_source.sw = "
        + f"{audio_source.sw} instead of 2"
    )
    assert audio_source.channels == 1, (
        "Unexpected number of channels: audio_source.channels = "
        + f"{audio_source.channels} instead of 1"
    )
    assert audio_source.ch == 1, (
        "Unexpected number of channels: audio_source.ch = "
        + f"{audio_source.ch} instead of 1"
    )


def test_get_audio_source_alias_prams():
    audio_source = get_audio_source(b"0" * 1600, sr=16000, sw=2, ch=1)
    assert audio_source.sampling_rate == 16000, (
        "Unexpected sampling rate: audio_source.sampling_rate = "
        + f"{audio_source.sampling_rate} instead of 16000"
    )
    assert audio_source.sr == 16000, (
        "Unexpected sampling rate: audio_source.sr = "
        + f"{audio_source.sr} instead of 16000"
    )
    assert audio_source.sample_width == 2, (
        "Unexpected sample width: audio_source.sample_width = "
        + f"{audio_source.sample_width} instead of 2"
    )
    assert audio_source.sw == 2, (
        "Unexpected sample width: audio_source.sw = "
        + f"{audio_source.sw} instead of 2"
    )
    assert audio_source.channels == 1, (
        "Unexpected number of channels: audio_source.channels = "
        + f"{audio_source.channels} instead of 1"
    )
    assert audio_source.ch == 1, (
        "Unexpected number of channels: audio_source.ch = "
        + f"{audio_source.ch} instead of 1"
    )
